import { createLoadEntitiesInitialState, createLoadEntitiesTraitReducer, } from './load-entities.trait.reducer';
import { loadEntitiesTraitKey, } from './load-entities.model';
import { createLoadEntitiesTraitMutators } from './load-entities.mutators';
import { createTraitFactory } from 'ngrx-traits';
import { createEntityAdapter } from '@ngrx/entity';
import { createLoadEntitiesTraitActions } from './load-entities.trait.actions';
import { createLoadEntitiesTraitSelectors } from './load-entities.trait.selectors';
/**
 * Generates the ngrx code needed to load a list of entities from the backend
 * with a fetch, fetchSuccess, fetchFail actions, and selectors to query the
 * entities like selectAll, selectIds, selectEntities, and its progress loading
 * with isLoading,isSuccess, isFail. This trait is the base for all other traits related
 * to a list of entities, the other will call fetch when needing data.
 * @param traitConfig - Config object fot the trait factory
 * @param traitConfig.selectId - Function that returns the id of an entity
 * @param traitConfig.sortComparer - Default sort function for to @ngrx/entity EntityAdapter
 *
 * @example
 * // The following trait config
 *
 * export interface TestState
 * extends EntityAndStatusState<Todo>{}
 *
 *    const traits = createEntityFeatureFactory(
 *      addLoadEntities<Todo>(),
 *    )({
 *      actionsGroupKey: '[Todos]',
 *      featureSelector: createFeatureSelector<TestState>>(
 *        'todos',
 *      ),
 *    });
 * // will generate the actions and selectors
 * traits.actions.fetch()
 * traits.actions.fetchSuccess({entities: todos})
 * traits.actions.fetchFail();
 * traits.selectors.selectAll
 * traits.selectors.selectEntities
 * traits.selectors.selectIds
 * traits.selectors.selectTotal
 * traits.selectors.isLoading
 * traits.selectors.isSuccess
 * traits.selectors.isFail
 */
export function addLoadEntities(traitConfig) {
    const adapter = createEntityAdapter(traitConfig);
    return createTraitFactory({
        key: loadEntitiesTraitKey,
        config: Object.assign(Object.assign({}, traitConfig), { adapter }),
        actions: ({ actionsGroupKey }) => createLoadEntitiesTraitActions(actionsGroupKey),
        selectors: ({ allConfigs }) => createLoadEntitiesTraitSelectors(allConfigs),
        mutators: ({ allConfigs }) => createLoadEntitiesTraitMutators(allConfigs),
        initialState: ({ previousInitialState, allConfigs, }) => createLoadEntitiesInitialState(previousInitialState, allConfigs),
        reducer: ({ initialState, allActions, allMutators, allConfigs }) => createLoadEntitiesTraitReducer(initialState, allActions, allMutators, allConfigs),
    });
}
//# sourceMappingURL=load-entities.trait.js.map