import { createMultiSelectionTraitActions } from './multi-selection.trait.actions';
import { createMultiSelectionTraitSelectors } from './multi-selection.trait.selectors';
import { loadEntitiesTraitKey, } from '../load-entities/load-entities.model';
import { createTraitFactory } from 'ngrx-traits';
import { createMultiSelectionInitialState, createMultiSelectionTraitReducer, } from './multi-selection.trait.reducer';
import { createMultiSelectionTraitMutators } from './multi-selection.trait.mutators';
/**
 * Generates ngrx code to add multi selection to a list
 *
 * @example
 * // The following trait config
 *
 * export interface TestState
 * extends EntityAndStatusState<Todo>,MultipleSelectionState{}
 *
 *    const traits = createEntityFeatureFactory(
 *      addLoadEntities<Todo>(),
 *      addMultiSelection<Todo>()
 *    )({
 *      actionsGroupKey: '[Todos]',
 *      featureSelector: createFeatureSelector<TestState>>(
 *        'todos',
 *      ),
 *    });
 * // will generate  the following actions and selectors, plus the ones generated by other traits
 * traits.actions.multiSelect({id})
 * traits.actions.multiDeselect({id})
 * traits.actions.multiToggleSelect({id})
 * traits.actions.toggleSelectAll()
 * traits.actions.multiClearSelection()
 *
 * traits.selectors.selectIdSelected()
 * traits.selectors.selectEntitySelected()
 */
export function addMultiSelection() {
    return createTraitFactory({
        key: 'multiSelection',
        depends: [loadEntitiesTraitKey],
        actions: ({ actionsGroupKey }) => createMultiSelectionTraitActions(actionsGroupKey),
        selectors: ({ previousSelectors }) => createMultiSelectionTraitSelectors(previousSelectors),
        initialState: ({ previousInitialState }) => createMultiSelectionInitialState(previousInitialState),
        mutators: ({ allSelectors }) => createMultiSelectionTraitMutators(allSelectors),
        reducer: ({ initialState, allActions, allMutators, allConfigs }) => createMultiSelectionTraitReducer(initialState, allActions, allMutators, allConfigs),
    });
}
//# sourceMappingURL=multi-selection.trait.js.map