import { createTraitFactory, } from 'ngrx-traits';
import { createSingleSelectionTraitActions } from './single-selection.trait.actions';
import { createSingleSelectionTraitSelectors } from './single-selection.trait.selectors';
import { createSingleSelectionInitialState, createSingleSelectionTraitReducer, } from './single-selection.trait.reducer';
import { createSingleSelectionTraitMutators } from './single-selection.trait.mutators';
import { loadEntitiesTraitKey, } from '../load-entities/load-entities.model';
/**
 * Generates ngrx code to add single selection to a list
 * @param config
 * @param config.selectedId - Default selected id
 *
 * @example
 * // The following trait config
 *
 * export interface TestState
 * extends EntityAndStatusState<Todo>,SingleSelectionState{}
 *
 *    const traits = createEntityFeatureFactory(
 *      addLoadEntities<Todo>(),
 *      addSingleSelection<Todo>()
 *    )({
 *      actionsGroupKey: '[Todos]',
 *      featureSelector: createFeatureSelector<TestState>>(
 *        'todos',
 *      ),
 *    });
 * // will generate  the following actions and selectors, plus the ones generated by other traits
 * traits.actions.select({id})
 * traits.actions.deselect()
 * traits.actions.toggleSelect({id})
 * traits.selectors.selectIdSelected()
 * traits.selectors.selectEntitySelected()
 */
export function addSingleSelection(config) {
    return createTraitFactory({
        key: 'singleSelection',
        depends: [loadEntitiesTraitKey],
        config,
        actions: ({ actionsGroupKey }) => createSingleSelectionTraitActions(actionsGroupKey),
        selectors: () => createSingleSelectionTraitSelectors(),
        mutators: () => createSingleSelectionTraitMutators(),
        initialState: ({ previousInitialState, allConfigs, }) => createSingleSelectionInitialState(previousInitialState, allConfigs),
        reducer: ({ initialState, allActions, allMutators, allConfigs }) => createSingleSelectionTraitReducer(initialState, allActions, allMutators, allConfigs),
    });
}
//# sourceMappingURL=single-selection.trait.js.map